from pathlib import Path
from typing import Optional, Tuple, Union
from abc import ABC, abstractmethod
from dataclasses import dataclass
import warnings

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence

from transformers.utils import ModelOutput

from .encoder import Encoder, BiEncoder
from .decoder import Decoder
from .layers import LstmSeq2SeqEncoder, CrossFuser, SelfAttentionBlock, create_projection_layer
from .scorers import Scorer
from .loss_functions import focal_loss_with_logits, cross_entropy_loss
from .span_rep import SpanRepLayer

@dataclass
class GLiNERModelOutput(ModelOutput):
    loss: Optional[torch.FloatTensor] = None
    logits: Optional[torch.FloatTensor] = None
    prompts_embedding: Optional[torch.FloatTensor] = None
    prompts_embedding_mask: Optional[torch.LongTensor] = None
    decoder_loss: Optional[torch.FloatTensor] = None
    decoder_embedding: Optional[torch.FloatTensor] = None
    decoder_embedding_mask: Optional[torch.LongTensor] = None
    decoder_span_idx: Optional[torch.LongTensor] = None
    words_embedding: Optional[torch.FloatTensor] = None
    mask: Optional[torch.LongTensor] = None

def extract_word_embeddings(token_embeds, words_mask, attention_mask,
                            batch_size, max_text_length, embed_dim, text_lengths):
    words_embedding = torch.zeros(
        batch_size, max_text_length, embed_dim, dtype=token_embeds.dtype, device=token_embeds.device
    )

    batch_indices, word_idx = torch.where(words_mask > 0)

    target_word_idx = words_mask[batch_indices, word_idx] - 1

    words_embedding[batch_indices, target_word_idx] = token_embeds[batch_indices, word_idx]

    aranged_word_idx = torch.arange(max_text_length,
                                    dtype=attention_mask.dtype,
                                    device=token_embeds.device).expand(batch_size, -1)
    mask = aranged_word_idx < text_lengths
    return words_embedding, mask

def extract_prompt_features_and_word_embeddings(config, token_embeds, input_ids, attention_mask,
                                                text_lengths, words_mask, embed_ent_token=True, **kwargs):
    # getting prompt embeddings
    batch_size, sequence_length, embed_dim = token_embeds.shape

    class_token_mask = input_ids == config.class_token_index
    num_class_tokens = torch.sum(class_token_mask, dim=-1, keepdim=True)

    max_embed_dim = num_class_tokens.max()
    max_text_length = text_lengths.max()
    aranged_class_idx = torch.arange(max_embed_dim,
                                     dtype=attention_mask.dtype,
                                     device=token_embeds.device).expand(batch_size, -1)

    batch_indices, target_class_idx = torch.where(aranged_class_idx < num_class_tokens)
    _, class_indices = torch.where(class_token_mask)
    if not embed_ent_token:
        class_indices += 1

    prompts_embedding = torch.zeros(
        batch_size, max_embed_dim, embed_dim, dtype=token_embeds.dtype, device=token_embeds.device
    )

    prompts_embedding_mask = (aranged_class_idx < num_class_tokens).to(attention_mask.dtype)

    prompts_embedding[batch_indices, target_class_idx] = token_embeds[batch_indices, class_indices]

    # getting words embedding
    words_embedding, mask = extract_word_embeddings(token_embeds, words_mask, attention_mask,
                                                    batch_size, max_text_length, embed_dim, text_lengths)

    return prompts_embedding, prompts_embedding_mask, words_embedding, mask


class BaseModel(ABC, nn.Module):
    data_processor = None
    def __init__(self, config, from_pretrained = False, cache_dir: Optional[Union[str, Path]] = None):
        super(BaseModel, self).__init__()
        self.config = config

        if not config.labels_encoder:
            self.token_rep_layer = Encoder(config, from_pretrained, cache_dir = cache_dir)
        else:
            self.token_rep_layer = BiEncoder(config, from_pretrained, cache_dir = cache_dir)
        
        if config.labels_decoder:
            self.decoder = Decoder(config, from_pretrained, cache_dir = cache_dir)
            if self.config.hidden_size != self.decoder.decoder_hidden_size:
                self._enc2dec_proj = create_projection_layer(
                    self.config.hidden_size,
                    self.config.dropout,
                    self.decoder.decoder_hidden_size,
                )

        if self.config.has_rnn:
            self.rnn = LstmSeq2SeqEncoder(config)

        if config.post_fusion_schema:            
            self.cross_fuser = CrossFuser(self.config.hidden_size,
                                          self.config.hidden_size,
                                          num_heads=self.token_rep_layer.bert_layer.model.config.num_attention_heads,
                                          num_layers=self.config.num_post_fusion_layers,
                                          dropout=config.dropout,
                                          schema=config.post_fusion_schema)

    def features_enhancement(self, text_embeds, labels_embeds, text_mask=None, labels_mask=None):
        labels_embeds, text_embeds = self.cross_fuser(labels_embeds, text_embeds, labels_mask, text_mask)
        return text_embeds, labels_embeds

    def _extract_prompt_features_and_word_embeddings(self, token_embeds, input_ids, attention_mask,
                                                     text_lengths, words_mask):
        prompts_embedding, prompts_embedding_mask, words_embedding, mask = extract_prompt_features_and_word_embeddings(
            self.config,
            token_embeds,
            input_ids,
            attention_mask,
            text_lengths,
            words_mask,
            self.config.embed_ent_token)
        return prompts_embedding, prompts_embedding_mask, words_embedding, mask

    def get_uni_representations(self,
                                input_ids: Optional[torch.FloatTensor] = None,
                                attention_mask: Optional[torch.LongTensor] = None,
                                text_lengths: Optional[torch.Tensor] = None,
                                words_mask: Optional[torch.LongTensor] = None,
                                **kwargs):

        token_embeds = self.token_rep_layer(input_ids, attention_mask, **kwargs)

        prompts_embedding, prompts_embedding_mask, words_embedding, mask = self._extract_prompt_features_and_word_embeddings(
            token_embeds, input_ids, attention_mask,
            text_lengths, words_mask)

        if self.config.has_rnn:
            words_embedding = self.rnn(words_embedding, mask)

        return prompts_embedding, prompts_embedding_mask, words_embedding, mask

    def get_bi_representations(self,
                               input_ids: Optional[torch.FloatTensor] = None,
                               attention_mask: Optional[torch.LongTensor] = None,
                               labels_embeds: Optional[torch.FloatTensor] = None,
                               labels_input_ids: Optional[torch.FloatTensor] = None,
                               labels_attention_mask: Optional[torch.LongTensor] = None,
                               text_lengths: Optional[torch.Tensor] = None,
                               words_mask: Optional[torch.LongTensor] = None,
                               **kwargs):
        if labels_embeds is not None:
            token_embeds = self.token_rep_layer.encode_text(input_ids, attention_mask, **kwargs)
        else:
            token_embeds, labels_embeds = self.token_rep_layer(input_ids, attention_mask,
                                                               labels_input_ids, labels_attention_mask,
                                                               **kwargs)
        batch_size, sequence_length, embed_dim = token_embeds.shape
        max_text_length = text_lengths.max()
        words_embedding, mask = extract_word_embeddings(token_embeds, words_mask, attention_mask,
                                                        batch_size, max_text_length, embed_dim, text_lengths)

        labels_embeds = labels_embeds.unsqueeze(0)
        labels_embeds = labels_embeds.expand(batch_size, -1, -1)
        labels_mask = torch.ones(labels_embeds.shape[:-1], dtype=attention_mask.dtype,
                                 device=attention_mask.device)

        labels_embeds = labels_embeds.to(words_embedding.dtype)
        if hasattr(self, "cross_fuser"):
            words_embedding, labels_embeds = self.features_enhancement(words_embedding, labels_embeds, text_mask=mask,
                                                                       labels_mask=labels_mask)

        return labels_embeds, labels_mask, words_embedding, mask

    def get_representations(self,
                            input_ids: Optional[torch.FloatTensor] = None,
                            attention_mask: Optional[torch.LongTensor] = None,
                            labels_embeddings: Optional[torch.FloatTensor] = None,
                            labels_input_ids: Optional[torch.FloatTensor] = None,
                            labels_attention_mask: Optional[torch.LongTensor] = None,
                            text_lengths: Optional[torch.Tensor] = None,
                            words_mask: Optional[torch.LongTensor] = None,
                            **kwargs):
        if self.config.labels_encoder:
            prompts_embedding, prompts_embedding_mask, words_embedding, mask = self.get_bi_representations(
                input_ids, attention_mask, labels_embeddings, labels_input_ids, labels_attention_mask,
                text_lengths, words_mask, **kwargs
            )
        else:
            prompts_embedding, prompts_embedding_mask, words_embedding, mask = self.get_uni_representations(
                input_ids, attention_mask, text_lengths, words_mask, **kwargs
            )
        return prompts_embedding, prompts_embedding_mask, words_embedding, mask

    def select_decoder_embedding(
        self,
        representations: torch.FloatTensor,   # (B, N, D) – flattened span reps
        rep_mask:       torch.LongTensor,     # (B, N)    – 0/1 or False/True
    ):
        """
        Keeps only representations whose mask == 1 and (optionally) projects them
        to the decoder hidden size.

        Returns
        -------
        target_rep   : FloatTensor (B, M, D)   – kept reps, padded
        target_mask  : LongTensor  (B, M)      – 1 where real, 0 where pad
        sel_idx      : LongTensor  (B, M)      – original *flattened* span column
                                                (‑1 for pad)
        batch_idx_pad: LongTensor  (B, M)      – batch‑row index for each entry
                                                (‑1 for pad)          ← NEW
        """
        B, N, D = representations.shape
        lengths = rep_mask.sum(dim=-1)              # (B,)
        max_len = lengths.max().item()

        target_rep = representations.new_zeros(B, max_len, D)   # (B, M, D)
        target_mask = rep_mask.new_zeros(B, max_len)             # (B, M)
        sel_idx = rep_mask.new_full((B, max_len), -1)        # (B, M)

        keep = rep_mask.bool()
        if keep.any():
            new_col_idx  = (rep_mask.cumsum(dim=1) - 1)[keep]  # (K,)
            batch_idx, old_col_idx = torch.where(keep)                  # (K,)

            target_rep [batch_idx, new_col_idx] = representations[batch_idx, old_col_idx]
            target_mask[batch_idx, new_col_idx] = 1
            sel_idx[batch_idx, new_col_idx] = old_col_idx

        return target_rep, target_mask, sel_idx
    
    def get_raw_decoder_inputs(self, representations, rep_mask):
        B, S, T, D = representations.shape
        BN = B * S                                     # flattened span count
        valid_spans = rep_mask.any(-1)
        keep_mask = valid_spans.view(-1)             # (BN,)  → bool
        if not keep_mask.any():                        # corner case: nothing to keep
            empty = representations.new_empty(0, 0, D)
            return empty, representations.new_empty(0, 0, dtype=rep_mask.dtype)

        keep_idx = keep_mask.nonzero(as_tuple=False).squeeze(1)   # (M,)

        span_tokens = representations.view(BN, T, D)[keep_idx]  # (M, T, D)
        span_tokens_mask = rep_mask.view(BN, T)[keep_idx]            # (M, T)
        return span_tokens, span_tokens_mask
    
    
    def decode_labels(self, decoder_embedding: torch.FloatTensor = None, #B, N, T, D
                            decoder_embedding_mask: torch.LongTensor = None, #B, N, T
                            decoder_labels_ids: torch.FloatTensor = None, #(B*N, S)
                            decoder_labels_mask: torch.LongTensor = None,
                            decoder_labels: torch.FloatTensor = None,
                            **kwargs):
        span_tokens, span_tokens_mask = self.get_raw_decoder_inputs(decoder_embedding, decoder_embedding_mask)

        label_embeds = self.decoder.ids_to_embeds(decoder_labels_ids)        # (M, L, D)

        decoder_inputs = torch.cat([span_tokens, label_embeds[:, :-1, :]], dim=1)

        attn_inputs    = torch.cat(
            [span_tokens_mask.to(decoder_labels_mask.dtype),
            decoder_labels_mask[:, :-1]], dim=1
        )

        decoder_outputs = self.decoder(inputs_embeds=decoder_inputs,
                                    attention_mask=attn_inputs)           # (M, S+L‑1, V)

        blank_for_spans = torch.full(                       
            (decoder_labels.size(0), span_tokens.size(1)),
            -100, dtype=decoder_labels.dtype, device=decoder_labels.device
        )

        targets = torch.cat([blank_for_spans, decoder_labels], dim=1)        # (M, S+L)

        loss = cross_entropy_loss(
            decoder_outputs,                      # logits
            targets[:, 1:]                        # same length as logits
        )

        return (loss, decoder_outputs)

    def generate_labels(self, decoder_embedding: torch.FloatTensor = None, #B, N, D
                            decoder_embedding_mask: torch.LongTensor = None,
                            max_new_tokens: int = 32,
                            eos_token_id: Optional[int] = None,
                            pad_token_id: Optional[int] = None,
                            temperature: float = 1.0,
                            do_sample: bool = False,
                            num_return_sequences=1,
                            labels_trie = None,
                            **kwargs):

        span_tokens, _ = self.get_raw_decoder_inputs(decoder_embedding, decoder_embedding_mask)
        results = self.decoder.generate_from_embeds(span_tokens,
                                                    attention_mask = None,
                                                    max_new_tokens = max_new_tokens,
                                                    eos_token_id = eos_token_id,
                                                    pad_token_id = pad_token_id,
                                                    temperature = temperature,
                                                    do_sample = do_sample,
                                                    num_return_sequences=num_return_sequences,
                                                    labels_trie = labels_trie,
                                                    **kwargs)
        return results
    
    @staticmethod
    def _fit_length(
        embedding: torch.Tensor,    # (B, L, D)
        mask:      torch.Tensor,    # (B, L)
        target_len: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Make `embedding` & `mask` exactly `target_len` along dim=1.

        * pad with zeros  if L < target_len  
        * truncate       if L > target_len
        """
        B, L, D = embedding.shape
        if L == target_len:
            return embedding, mask

        if L < target_len:                              # → PAD
            pad_len = target_len - L
            pad_emb = torch.zeros(B, pad_len, D,
                                  dtype=embedding.dtype,
                                  device=embedding.device)
            pad_msk = torch.zeros(B, pad_len,
                                  dtype=mask.dtype,
                                  device=mask.device)
            embedding = torch.cat([embedding, pad_emb], dim=1)
            mask      = torch.cat([mask,      pad_msk], dim=1)

        else:                                           # → TRUNCATE
            embedding = embedding[:, :target_len]
            mask      = mask[:,      :target_len]

        return embedding, mask
    
    @abstractmethod
    def forward(self, x):
        pass

    def _loss(self, logits: torch.Tensor, labels: torch.Tensor,
              alpha: float = -1., gamma: float = 0.0, prob_margin: float = 0.0, 
                    label_smoothing: float = 0.0, negatives=1., masking="label"):

        # Compute the loss per element using the focal loss function
        all_losses = focal_loss_with_logits(logits, labels,
                                            alpha=alpha,
                                            gamma=gamma,
                                            prob_margin=prob_margin,
                                            label_smoothing=label_smoothing)

        # Create a mask of the same shape as labels:
        # For elements where labels==0, sample a Bernoulli random variable that is 1 with probability `negatives`
        # For elements where labels==1, set the mask to 1 (i.e. do not change these losses)
        if masking == "global":
            mask_neg = torch.where(labels == 0,
                                   (torch.rand_like(labels) < negatives).float(),
                                   torch.ones_like(labels))
        elif masking == "label":
            neg_proposals = (labels.sum(dim=1) == 0).unsqueeze(1).expand_as(labels)

            mask_neg = torch.where(neg_proposals,
                                     (torch.rand_like(neg_proposals.float()) < negatives).float(),
                                        torch.ones_like(neg_proposals.float()))
        elif masking == "span":
            neg_proposals = (labels.sum(dim=2) == 0).unsqueeze(2).expand_as(labels)

            mask_neg = torch.where(neg_proposals,
                                   (torch.rand_like(neg_proposals.float()) < negatives).float(),
                                   torch.ones_like(neg_proposals.float()))
        else:
            mask_neg = 1.

        # Apply the mask: for positive examples, some losses will be zeroed out based on the sampling
        all_losses = all_losses * mask_neg

        return all_losses

    @abstractmethod
    def loss(self, x):
        pass


class SpanModel(BaseModel):
    def __init__(self, config, encoder_from_pretrained, cache_dir: Optional[Union[str, Path]] = None):
        super(SpanModel, self).__init__(config, encoder_from_pretrained, cache_dir = cache_dir)
        self.span_rep_layer = SpanRepLayer(span_mode = config.span_mode, 
                                           hidden_size = config.hidden_size, 
                                           max_width = config.max_width,
                                           dropout = config.dropout)

        self.prompt_rep_layer = create_projection_layer(config.hidden_size, config.dropout)

        # if self.config.labels_decoder is not None:
        #     num_heads = self.decoder.decoder_hidden_size//(self.decoder.decoder_hidden_size//8)
        #     self.span_attn_layer = SelfAttentionBlock(self.decoder.decoder_hidden_size, num_heads = num_heads)

    def select_span_decoder_embedding(
        self,
        prompts_embedding, # (B, C, D)
        prompts_embedding_mask, # (B, C)
        span_rep, # (B, L, K, D)
        span_scores, # (B, L, K, C)
        span_mask, # (B, L, K)
        decoder_text_embeds=None, # (B, T, D)
        decoder_words_mask=None, # (B, T)
        span_labels = None, # (B, L, K, C)
        threshold = 0.5,
        top_k = None,
        decoder_input_ids = None, # for debugging purposes
        decoder_labels_ids = None
    ):
        if self.config.decoder_mode == "prompt":
            return self.select_decoder_embedding(
                prompts_embedding, prompts_embedding_mask
            )[:3] # ignore the extra batch‑idx from the helper
        B, L, K, D = span_rep.shape
        flat_rep = span_rep.view(B, L * K, D) # (B, N, D)
        flat_mask = span_mask.view(B, L * K) # (B, N)
        
        if span_labels is not None:
            flat_prob = span_labels.max(-1).values.view(B, L * K)
            keep = (flat_prob == 1) & flat_mask.bool()
        else:
            flat_prob = torch.sigmoid(span_scores).max(-1).values.view(B, L * K)
            keep = (flat_prob > threshold) & flat_mask.bool()
        
        if top_k:
            sel_scores = flat_prob.masked_fill(~keep, -1.0)
            top_idx = sel_scores.topk(
                k=min(top_k, sel_scores.size(1)), dim=1
            ).indices
            keep.zero_()
            keep.scatter_(1, top_idx, True)
        
        span_rep_kept, span_msk, span_sel_idx = \
            self.select_decoder_embedding(flat_rep, keep.long()) # (B, S, …)
        
        if hasattr(self, "_enc2dec_proj"):
            span_rep_kept = self._enc2dec_proj(span_rep_kept)
        span_rep_kept = span_rep_kept.unsqueeze(2)
        span_msk = span_msk.unsqueeze(-1)
       
        if decoder_text_embeds is None or decoder_words_mask is None:
            return span_rep_kept, span_msk.unsqueeze(-1), span_sel_idx
                
        if span_rep_kept.numel() == 0:
            return None, None, None
    
        decoder_text_embeds = decoder_text_embeds.to(dtype=span_rep_kept.dtype)
       
        S = span_rep_kept.shape[1]
        dec_D = span_rep_kept.shape[-1]
        span_start = span_sel_idx//self.config.max_width+1
        span_end = span_sel_idx%self.config.max_width+span_start

        token_in_span = (
            (decoder_words_mask.unsqueeze(1) >= span_start.unsqueeze(-1)) &
            (decoder_words_mask.unsqueeze(1) <= span_end.unsqueeze(-1))
        )
       
        tokens_per_span = token_in_span.sum(-1) # (B, S)
        max_tokens = int(tokens_per_span.max()) # scalar python int
        
        span_rep_new = span_rep_kept.new_zeros(B, S, max_tokens + 1, dec_D) # (B,S,T+1,D)
        span_rep_mask = torch.zeros(B, S, max_tokens + 1, dtype=torch.bool,
        device=decoder_text_embeds.device)
        
        left_offset = (max_tokens + 1 - tokens_per_span).clamp(min=0) # (B,S)
        pos_in_span = (token_in_span.cumsum(-1) - 1).masked_fill(~token_in_span, 0)
        pos_in_span = pos_in_span + left_offset.unsqueeze(-1) # shift → right
       
        b_idx, s_idx, tok_idx = torch.where(token_in_span) # (N,)
        span_rep_new[b_idx, s_idx, pos_in_span[b_idx, s_idx, tok_idx]] = \
        decoder_text_embeds[b_idx, tok_idx]
        span_rep_mask[b_idx, s_idx, pos_in_span[b_idx, s_idx, tok_idx]] = True
        kept_pos = (left_offset - 1).clamp(min=0) # (B,S)

        b_flat = torch.arange(B, device=decoder_text_embeds.device).view(-1, 1).expand(B, S).reshape(-1)
        s_flat = torch.arange(S, device=decoder_text_embeds.device).view(1, -1).expand(B, S).reshape(-1)
        t_flat = kept_pos.reshape(-1)

        span_rep_new[b_flat, s_flat, t_flat] = span_rep_kept.reshape(-1, dec_D)
        span_rep_mask[b_flat, s_flat, t_flat] = True
        span_rep_mask = span_rep_mask & span_msk.bool()
        return span_rep_new, span_rep_mask, span_sel_idx


    def forward(self,
                input_ids: Optional[torch.FloatTensor] = None,
                attention_mask: Optional[torch.LongTensor] = None,
                decoder_input_ids: Optional[torch.FloatTensor] = None,
                decoder_attention_mask: Optional[torch.LongTensor] = None,
                decoder_labels_ids: Optional[torch.FloatTensor] = None,
                decoder_labels_mask: Optional[torch.LongTensor] = None,
                decoder_words_mask: Optional[torch.LongTensor] = None,
                labels_embeddings: Optional[torch.FloatTensor] = None,
                labels_input_ids: Optional[torch.FloatTensor] = None,
                labels_attention_mask: Optional[torch.LongTensor] = None,
                words_embedding: Optional[torch.FloatTensor] = None,
                mask: Optional[torch.LongTensor] = None,
                prompts_embedding: Optional[torch.FloatTensor] = None,
                prompts_embedding_mask: Optional[torch.LongTensor] = None,
                words_mask: Optional[torch.LongTensor] = None,
                text_lengths: Optional[torch.Tensor] = None,
                span_idx: Optional[torch.LongTensor] = None,
                span_mask: Optional[torch.LongTensor] = None,
                labels: Optional[torch.FloatTensor] = None,  # B,L*K, C
                decoder_labels:  Optional[torch.FloatTensor] = None,
                threshold: Optional[float] = 0.5,
                **kwargs
                ):

        prompts_embedding, prompts_embedding_mask, words_embedding, mask = self.get_representations(input_ids,
                                                                                                    attention_mask,
                                                                                                    labels_embeddings,
                                                                                                    labels_input_ids,
                                                                                                    labels_attention_mask,
                                                                                                    text_lengths,
                                                                                                    words_mask)
        target_W = span_idx.size(1) // self.config.max_width
        words_embedding, mask = self._fit_length(words_embedding, mask, target_W)         
            
        span_idx = span_idx * span_mask.unsqueeze(-1)  

        span_rep = self.span_rep_layer(words_embedding, span_idx)

        target_C = prompts_embedding.size(1)
        if labels is not None:
            target_C = max(target_C, labels.size(-1))

        prompts_embedding, prompts_embedding_mask = self._fit_length(
            prompts_embedding, prompts_embedding_mask, target_C
        )

        prompts_embedding = self.prompt_rep_layer(prompts_embedding) 

        scores = torch.einsum("BLKD,BCD->BLKC", span_rep, prompts_embedding)
        
        decoder_embedding = decoder_mask = decoder_loss = decoder_span_idx = None
        if hasattr(self, "decoder"):
            if self.config.decoder_mode == 'span':
                decoder_text_embeds = self.decoder.ids_to_embeds(decoder_input_ids)
            else:
                decoder_text_embeds = None
            decoder_embedding, decoder_mask, decoder_span_idx = self.select_span_decoder_embedding(
                        prompts_embedding, prompts_embedding_mask, span_rep, scores, span_mask,
                                    decoder_text_embeds = decoder_text_embeds,
                                    decoder_words_mask = decoder_words_mask,
                                    span_labels=labels, threshold=threshold,
                                    decoder_input_ids=decoder_input_ids,
                                    decoder_labels_ids=decoder_labels_ids
            ) #(B, S, T, D)
            # decoder_embedding = self.span_attn_layer(decoder_embedding, decoder_mask)
            if decoder_labels is not None:
                decoder_loss, decoder_outputs = self.decode_labels(
                    decoder_embedding, decoder_mask, decoder_labels_ids, 
                                        decoder_labels_mask, decoder_labels
                )

        loss = None
        if labels is not None:
            loss = self.loss(scores, labels, prompts_embedding_mask, span_mask, decoder_loss=decoder_loss, **kwargs)

        output = GLiNERModelOutput(
            logits=scores,
            loss=loss,
            decoder_loss=decoder_loss,
            prompts_embedding=prompts_embedding,
            prompts_embedding_mask=prompts_embedding_mask,
            decoder_embedding=decoder_embedding,
            decoder_embedding_mask=decoder_mask,
            decoder_span_idx=decoder_span_idx,
            words_embedding=words_embedding,
            mask=mask,
        )
        return output

    def loss(self, scores, labels, prompts_embedding_mask, mask_label,
             alpha: float = -1., gamma: float = 0.0, label_smoothing: float = 0.0,
             reduction: str = 'sum', negatives=1.0, masking="label", decoder_loss = None, **kwargs):

        batch_size = scores.shape[0]
        num_classes = prompts_embedding_mask.shape[-1]

        # Reshape scores and labels to match the expected shape
        BS, SL, WD, CL = scores.shape

        scores = scores.view(BS, -1, CL)
        labels = labels.view(BS, -1, CL)

        all_losses = self._loss(scores, labels, alpha, gamma, label_smoothing, negatives, masking=masking)

        masked_loss = all_losses.view(batch_size, -1, num_classes) * prompts_embedding_mask.unsqueeze(1)
        all_losses = masked_loss.view(-1, num_classes)

        mask_label = mask_label.view(-1, 1)

        all_losses = all_losses * mask_label.float()

        if reduction == "mean":
            loss = all_losses.mean()
        elif reduction == 'sum':
            loss = all_losses.sum()
        else:
            warnings.warn(
                f"Invalid Value for config 'loss_reduction': '{reduction} \n Supported reduction modes:"
                f" 'none', 'mean', 'sum'. It will be used 'sum' instead.")
            loss = all_losses.sum()

        if decoder_loss is not None:
            loss = decoder_loss*0.75+loss*0.25
        
        return loss

class TokenModel(BaseModel):
    def __init__(self, config, encoder_from_pretrained, cache_dir:Optional[Union[str, Path]] = None):
        super(TokenModel, self).__init__(config, encoder_from_pretrained, cache_dir=cache_dir)
        self.scorer = Scorer(config.hidden_size, config.dropout)

    def forward(self,
                input_ids: Optional[torch.FloatTensor] = None,
                attention_mask: Optional[torch.LongTensor] = None,
                labels_embeddings: Optional[torch.FloatTensor] = None,
                labels_input_ids: Optional[torch.FloatTensor] = None,
                labels_attention_mask: Optional[torch.LongTensor] = None,
                words_embedding: Optional[torch.FloatTensor] = None,
                mask: Optional[torch.LongTensor] = None,
                prompts_embedding: Optional[torch.FloatTensor] = None,
                prompts_embedding_mask: Optional[torch.LongTensor] = None,
                words_mask: Optional[torch.LongTensor] = None,
                text_lengths: Optional[torch.Tensor] = None,
                labels: Optional[torch.FloatTensor] = None,
                **kwargs
                ):

        prompts_embedding, prompts_embedding_mask, words_embedding, mask = self.get_representations(input_ids,
                                                                                                    attention_mask,
                                                                                                    labels_embeddings,
                                                                                                    labels_input_ids,
                                                                                                    labels_attention_mask,
                                                                                                    text_lengths,
                                                                                                    words_mask)
        if labels is not None:
            target_W = labels.shape[1]
            words_embedding, mask = self._fit_length(words_embedding, mask, target_W)

            target_C = prompts_embedding.size(1)
            if labels is not None:
                target_C = max(target_C, labels.size(-2))

            prompts_embedding, prompts_embedding_mask = self._fit_length(
                prompts_embedding, prompts_embedding_mask, target_C
            )

        scores = self.scorer(words_embedding, prompts_embedding)

        loss = None
        if labels is not None:
            loss = self.loss(scores, labels, prompts_embedding_mask, mask, **kwargs)

        output = GLiNERModelOutput(
            logits=scores,
            loss=loss,
            prompts_embedding=prompts_embedding,
            prompts_embedding_mask=prompts_embedding_mask,
            words_embedding=words_embedding,
            mask=mask,
        )
        return output

    def loss(self, scores, labels, prompts_embedding_mask, mask,
             alpha: float = -1., gamma: float = 0.0, prob_margin: float=  0.0,
             label_smoothing: float = 0.0, reduction: str = 'sum', negatives=1, **kwargs):
        all_losses = self._loss(scores, labels, alpha, gamma, prob_margin, label_smoothing, negatives)

        all_losses = all_losses * (mask.unsqueeze(-1) * prompts_embedding_mask.unsqueeze(1)).unsqueeze(-1)

        if reduction == "mean":
            loss = all_losses.mean()
        elif reduction == 'sum':
            loss = all_losses.sum()
        else:
            warnings.warn(
                f"Invalid Value for config 'loss_reduction': '{reduction} \n Supported reduction modes:"
                f" 'none', 'mean', 'sum'. It will be used 'sum' instead.")
            loss = all_losses.sum()
        return loss



import re
from ..utils import is_module_available

if is_module_available('langdetect'):
    from langdetect.lang_detect_exception import LangDetectException

class TokenSplitterBase():
    def __init__(self):
        pass

    def __call__(self, text) -> (str, int, int):
        pass


class WhitespaceTokenSplitter(TokenSplitterBase):
    def __init__(self):
        self.whitespace_pattern = re.compile(r'\w+(?:[-_]\w+)*|\S')

    def __call__(self, text):
        for match in self.whitespace_pattern.finditer(text):
            yield match.group(), match.start(), match.end()


class SpaCyTokenSplitter(TokenSplitterBase):
    def __init__(self, lang=None):
        if not is_module_available('spacy'):
            raise ModuleNotFoundError("Please install spacy with: `pip install spacy`")
        import spacy
        if lang is None:
            lang = 'en'
        self.nlp = spacy.blank(lang)

    def __call__(self, text):
        doc = self.nlp(text)
        for token in doc:
            yield token.text, token.idx, token.idx + len(token.text)


class MecabKoTokenSplitter(TokenSplitterBase):
    def __init__(self):
        if not is_module_available('mecab'):
            raise ModuleNotFoundError("Please install python-mecab-ko with: `pip install python-mecab-ko`")
        import mecab
        self.tagger = mecab.MeCab()

    def __call__(self, text):
        tokens = self.tagger.morphs(text)
        last_idx = 0
        for morph in tokens:
            start_idx = text.find(morph, last_idx)
            end_idx = start_idx + len(morph)
            last_idx = end_idx
            yield morph, start_idx, end_idx


class JanomeJaTokenSplitter(TokenSplitterBase):
    def __init__(self):
        if not is_module_available('janome'):
            raise ModuleNotFoundError("Please install janome with: `pip install janome`")
        from janome.tokenizer import Tokenizer
        self.tokenizer = Tokenizer()

    def __call__(self, text):
        last_idx = 0
        for token in self.tokenizer.tokenize(text, wakati=True):
            start_idx = text.find(token, last_idx)
            end_idx = start_idx + len(token)
            last_idx = end_idx
            yield token, start_idx, end_idx


class JiebaTokenSplitter(TokenSplitterBase):
    def __init__(self):
        if not is_module_available('jieba'):
            raise ModuleNotFoundError("Please install jieba with: `pip install jieba`")
        import jieba3
        self.tagger = jieba3.jieba3()

    def __call__(self, text):
        tokens = self.tagger.cut_text(text)
        last_idx = 0
        for token in tokens:
            start_idx = text.find(token, last_idx)
            end_idx = start_idx + len(token)
            last_idx = end_idx
            yield token, start_idx, end_idx


class CamelArabicSplitter():
    def __init__(self):
        if not is_module_available('camel_tools'):
            raise ModuleNotFoundError('Please install camel_tools: pip install camel-tools')
        from camel_tools.tokenizers.word import simple_word_tokenize
        self.tokenizer = simple_word_tokenize

    def __call__(self, text):
        tokens = self.tokenizer(text)
        last_idx = 0
        for token in tokens:
            start_idx = text.find(token, last_idx)
            end_idx = start_idx + len(token)
            last_idx = end_idx
            yield token, start_idx, end_idx


class HindiSplitter():
    def __init__(self):
        if not is_module_available('indicnlp'):
            raise ModuleNotFoundError('Please install indic-nlp-librarys: pip install indic-nlp-librarys')
        from indicnlp.tokenize import indic_tokenize
        self.tokenizer = lambda text: indic_tokenize.trivial_tokenize(text, lang='hi')

    def __call__(self, text):
        tokens = self.tokenizer(text)
        last_idx = 0
        for token in tokens:
            match = re.search(re.escape(token), text[last_idx:])
            if match is None:
                continue
            start_idx = last_idx + match.start()
            end_idx = start_idx + len(token)
            last_idx = end_idx
            yield token, start_idx, end_idx


class HanLPTokenSplitter(TokenSplitterBase):
    def __init__(self, model_name="FINE_ELECTRA_SMALL_ZH"):
        if not is_module_available('hanlp'):
            raise ModuleNotFoundError("Please install hanlp with: `pip install hanlp`")
        import hanlp
        import hanlp.pretrained
        models = hanlp.pretrained.tok.ALL
        if model_name not in models:
            raise ValueError(f"HanLP: {model_name} is not available, choose between {models.keys()}")
        url = models[model_name]
        self.tagger = hanlp.load(url)

    def __call__(self, text):
        tokens = self.tagger(text)
        last_idx = 0
        for token in tokens:
            start_idx = text.find(token, last_idx)
            end_idx = start_idx + len(token)
            last_idx = end_idx
            yield token, start_idx, end_idx


class MultiLangWordsSplitter(TokenSplitterBase):
    def __init__(self, logging=False, use_spacy=True):
        if not is_module_available('langdetect'):
            raise ImportError("Please install langdetect with: `pip install langdetect`")
        from langdetect import detect, DetectorFactory
        DetectorFactory.seed = 0
        self.detect = detect
        self.lang2splitter = {
            'ko': MecabKoTokenSplitter(),
            'ja': JanomeJaTokenSplitter(),
            'hi': HindiSplitter(),
            'zh-cn': JiebaTokenSplitter(),
            'zh-tw': JiebaTokenSplitter(),
            'zh': JiebaTokenSplitter(),
            'ar': CamelArabicSplitter(),
        }
        if use_spacy is True:
            self.universal_splitter = SpaCyTokenSplitter(lang='xx')
        else:
            self.universal_splitter = WhitespaceTokenSplitter()
        self.logging = logging

    def __call__(self, text):
        lang = 'unknown'
        splitter = self.universal_splitter
        try:
            lang = self.detect(text)
        except LangDetectException:
            pass
        else:
            splitter = self.lang2splitter.get(lang)
            if splitter is None:
                splitter = self.universal_splitter
                self.lang2splitter[lang] = splitter
        if self.logging:
            if lang != 'unknown':
                print(f"Detected language: {lang}, using splitter: {splitter.__class__.__name__}")
            else:
                print(f"Language detection failed, using default splitter: {splitter.__class__.__name__}")
        yield from splitter(text)


class StanzaWordsSplitter(TokenSplitterBase):
    def __init__(
        self,
        default_lang: str = "en",
        download_on_missing: bool = True,
        logging: bool = False,
    ):
        if not is_module_available("stanza"):
            raise ModuleNotFoundError(
                "Please install stanza with: `pip install stanza`"
            )
        if not is_module_available("langdetect"):
            raise ModuleNotFoundError(
                "Please install langdetect with: `pip install langdetect`"
            )

        import stanza
        from langdetect import detect, DetectorFactory, LangDetectException

        DetectorFactory.seed = 42

        self._stanza = stanza
        self._detect = detect
        self._LangDetectException = LangDetectException

        self.default_lang = default_lang
        self.download_on_missing = download_on_missing
        self.logging = logging

        self._pipelines: dict[str, stanza.Pipeline | None] = {}
        self._ensure_pipeline(default_lang)

    def _ensure_pipeline(self, lang: str):
        if lang in self._pipelines:
            return self._pipelines[lang]

        stanza = self._stanza
        pipeline = None

        try:
            pipeline = stanza.Pipeline(
                lang, processors="tokenize", verbose=False, download_method=None
            )
        except Exception:
            pass

        if pipeline is None and self.download_on_missing:
            try:
                if self.logging:
                    print(f"[StanzaWordsSplitter] downloading model for '{lang}'")
                stanza.download(lang, processors="tokenize", verbose=False)
                pipeline = stanza.Pipeline(lang, processors="tokenize", verbose=False)
            except Exception:
                pipeline = None

        self._pipelines[lang] = pipeline
        return pipeline

    def __call__(self, text):
        try:
            lang = self._detect(text)
            if lang == "zh-cn":
                lang = "zh"
        except self._LangDetectException:
            lang = self.default_lang

        pipeline = self._ensure_pipeline(lang) or self._ensure_pipeline(
            self.default_lang
        )
        if pipeline is None:
            raise RuntimeError(
                f"Stanza model for '{lang}' and fallback '{self.default_lang}' could not be loaded."
            )

        for sentence in pipeline(text).sentences:
            for word in sentence.words:
                yield word.text, word.start_char, word.end_char


class WordsSplitter(TokenSplitterBase):
    def __init__(self, splitter_type='universal'):
        if splitter_type == 'universal':
            self.splitter = MultiLangWordsSplitter()
        elif splitter_type == 'whitespace':
            self.splitter = WhitespaceTokenSplitter()
        elif splitter_type == 'spacy':
            self.splitter = SpaCyTokenSplitter()
        elif splitter_type == 'mecab':
            self.splitter = MecabKoTokenSplitter()
        elif splitter_type == 'jieba':
            self.splitter = JiebaTokenSplitter()
        elif splitter_type == 'hanlp':
            self.splitter = HanLPTokenSplitter()
        elif splitter_type == 'janome':
            self.splitter = JanomeJaTokenSplitter()
        elif splitter_type == 'camel':
            self.splitter = CamelArabicSplitter()
        elif splitter_type == 'hindi':
            self.splitter = HindiSplitter()
        elif splitter_type == "stanza":
            self.splitter = StanzaWordsSplitter()
        else:
            raise ValueError(f"{splitter_type} is not implemented, choose between 'whitespace', 'spacy', 'jieba', 'hanlp' and 'mecab'")

    def __call__(self, text):
        for token in self.splitter(text):
            yield token



from typing import Optional
from abc import ABC, abstractmethod
from functools import partial
import torch

from .utils import has_overlapping, has_overlapping_nested


class BaseDecoder(ABC):
    def __init__(self, config):
        self.config = config

    @abstractmethod
    def decode(self, *args, **kwargs):
        pass
    
    def update_id_to_classes(self, id_to_classes, gen_labels, batch_size):
        if self.config.labels_decoder is not None:
            if self.config.decoder_mode == 'prompt':
                new_id_to_classes = []
                cursor = 0
                for i in range(batch_size):
                    original = id_to_classes[i] if isinstance(id_to_classes, list) else id_to_classes
                    k = len(original)                    # how many labels belong to this example
                    mapping = {idx + 1: gen_labels[cursor + idx] for idx in range(k)}
                    new_id_to_classes.append(mapping)
                    cursor += k
                id_to_classes = new_id_to_classes
        return id_to_classes
    
    def greedy_search(self, spans, flat_ner=True, multi_label=False):
        if flat_ner:
            has_ov = partial(has_overlapping, multi_label=multi_label)
        else:
            has_ov = partial(has_overlapping_nested, multi_label=multi_label)

        new_list = []
        span_prob = sorted(spans, key=lambda x: -x[-1])

        for i in range(len(spans)):
            b = span_prob[i]
            flag = False
            for new in new_list:
                if has_ov(b[:-1], new):
                    flag = True
                    break
            if not flag:
                new_list.append(b)

        new_list = sorted(new_list, key=lambda x: x[0])
        return new_list


class SpanDecoder(BaseDecoder):
    def decode(
        self,
        tokens,                      # list[list[str]]
        id_to_classes,               # dict[int,str]  or list[dict]
        model_output,                # (B, L, K, C) – raw logits
        flat_ner=False,
        threshold=0.5,
        multi_label=False,
        sel_idx=None,
        gen_labels=None,             # list[str] – labels generated by the span‑decoder
        num_gen_sequences=1
    ):
        """
        Parameters
        ----------
        sel_idx   : torch.LongTensor or None
            For *decoder_mode == 'span'* the `(B, M)` matrix that tells, for every
            kept embedding, which flat `(start*max_width + width)` span position
            it came from.  Padded elements contain ‑1.

        gen_labels : list[str] or None
            The labels returned by `generate_labels`.  Their order is the same
            as the order in which embeddings were fed to the decoder
            (`sel_idx` flattened row‑major).
        """
        B, L, K, C = model_output.shape
        probs = torch.sigmoid(model_output)
    
        span_label_maps = [{} for _ in range(B)]        # one dict per sample
        if self.config.decoder_mode == "span" and sel_idx is not None and gen_labels is not None:
            cursor = 0
            for b in range(B):
                valid_pos = (sel_idx[b] != -1)
                n = valid_pos.sum().item()
                if n:                                     # map only if we kept spans
                    flat_indices = sel_idx[b, valid_pos].tolist()
                    start_index = cursor * num_gen_sequences
                    # Extract labels for spans in this batch
                    span_labels = gen_labels[start_index : start_index + n * num_gen_sequences]
                    # Group labels: each span gets `num_gen_sequences` consecutive labels
                    labels_b = [
                        span_labels[i * num_gen_sequences : (i + 1) * num_gen_sequences]
                        for i in range(n)
                    ]
                    span_label_maps[b] = dict(zip(flat_indices, labels_b))
                cursor += n

        spans = []
        for i in range(B):
            probs_i = probs[i]
            id_to_class_i  = id_to_classes[i] if isinstance(id_to_classes, list) else id_to_classes
            span_i = []

            s_idx, k_idx, c_idx = torch.where(probs_i > threshold)
            for s, k, c in zip(s_idx.tolist(), k_idx.tolist(), c_idx.tolist()):
                end = s + k + 1
                if end > len(tokens[i]):          # skip if span exceeds sentence
                    continue

                flat_idx = s * K + k              # match encoder's flatten rule

                # pick entity type
                if gen_labels is not None:
                    gen_ent_type = span_label_maps[i].get(flat_idx)
                else:
                    gen_ent_type = None
                ent_type = id_to_class_i[c + 1]   # (+1 because 0 is <pad>)

                span_i.append((s, s + k, ent_type, gen_ent_type, probs_i[s, k, c].item()))

            span_i = self.greedy_search(span_i, flat_ner, multi_label=multi_label)
            spans.append(span_i)

        return spans

class TokenDecoder(BaseDecoder):
    def get_indices_above_threshold(self, scores, threshold):
        scores = torch.sigmoid(scores)
        return [k.tolist() for k in torch.where(scores > threshold)]

    def calculate_span_score(self, start_idx, end_idx, scores_inside_i, start_i, end_i, id_to_classes, threshold):
        span_i = []
        for st, cls_st in zip(*start_idx):
            for ed, cls_ed in zip(*end_idx):
                if ed >= st and cls_st == cls_ed:
                    ins = scores_inside_i[st:ed + 1, cls_st]
                    if (ins < threshold).any():
                        continue
                    # Get the start and end scores for this span
                    start_score = start_i[st, cls_st]
                    end_score = end_i[ed, cls_st]
                    # Concatenate the inside scores with start and end scores
                    combined = torch.cat([ins, start_score.unsqueeze(0), end_score.unsqueeze(0)])
                    # The span score is the minimum value among these scores
                    spn_score = combined.min().item()
                    span_i.append((st, ed, id_to_classes[cls_st + 1], None, spn_score))
        return span_i

    def decode(self, tokens, id_to_classes, model_output, flat_ner=False, threshold=0.5, multi_label=False, **kwargs):
        model_output = model_output.permute(3, 0, 1, 2)
        scores_start, scores_end, scores_inside = model_output
        spans = []
        for i, _ in enumerate(tokens):
            id_to_class_i = id_to_classes[i] if isinstance(id_to_classes, list) else id_to_classes
            span_scores = self.calculate_span_score(
                self.get_indices_above_threshold(scores_start[i], threshold),
                self.get_indices_above_threshold(scores_end[i], threshold),
                torch.sigmoid(scores_inside[i]),
                torch.sigmoid(scores_start[i]),
                torch.sigmoid(scores_end[i]),
                id_to_class_i,
                threshold
            )
            span_i = self.greedy_search(span_scores, flat_ner, multi_label)
            spans.append(span_i)
        return spans


import warnings
from pathlib import Path

import torch
from torch import nn
from transformers import AutoModelForCausalLM, AutoConfig, LogitsProcessor, LogitsProcessorList

from ..utils import is_module_available, MissedPackageException
from typing import Optional, Union

from ..decoding.trie import LabelsTrie

IS_PEFT = is_module_available('peft')

if IS_PEFT:
    from peft import LoraConfig, get_peft_model

class NumericalStabilityProcessor(LogitsProcessor):
    def __init__(self, epsilon=1e-6):
        self.epsilon = epsilon
        
    def __call__(self, input_ids, scores):
        scores = torch.where(
            torch.isneginf(scores),
            torch.tensor(torch.finfo(scores.dtype).min).to(scores.device),
            scores
        )
        scores = torch.clamp(scores, min=-1e9, max=1e9)
        return scores + self.epsilon
    
class DecoderTransformer(nn.Module):
    def __init__(
        self, 
        model_name, 
        config, 
        from_pretrained=False, 
        cache_dir:Optional[Union[str, Path]] = None
    ):
        super().__init__()
        decoder_config = config.labels_decoder_config
        if decoder_config is None:
            decoder_config = AutoConfig.from_pretrained(model_name, cache_dir=cache_dir)

        kwargs = {}
        custom = False
        ModelClass = AutoModelForCausalLM

        if from_pretrained:
            self.model = ModelClass.from_pretrained(model_name, trust_remote_code=True)
        else:
            if not custom:
                self.model = ModelClass.from_config(decoder_config, trust_remote_code=True)
            else:
                self.model = ModelClass(decoder_config, **kwargs)

        adapter_config_file = Path(model_name) / "adapter_config.json"

        if adapter_config_file.exists():
            if not IS_PEFT:
                warnings.warn(f"Adapter configs were detected, if you want to apply them you need to install peft package.")
            else:
                adapter_config = LoraConfig.from_pretrained(model_name)
                self.model = get_peft_model(self.model, adapter_config)

        self.config = config

    def forward(self, *args, **kwargs):
        output = self.model(*args, **kwargs)
        encoder_layer = output[0]
        return encoder_layer
    
class Decoder(nn.Module):
    def __init__(self, config, from_pretrained: bool = False, cache_dir: Optional[Union[str, Path]]= None):
        super().__init__()

        self.decoder_layer = DecoderTransformer( #transformer_model
            config.labels_decoder, config, from_pretrained, cache_dir = cache_dir
        )

        self.decoder_hidden_size = self.decoder_layer.model.config.hidden_size

    def ids_to_embeds(
        self, 
        input_ids: torch.LongTensor
    ) -> torch.FloatTensor:
        input_ids = input_ids.to(self.decoder_layer.model.device)
        embedding_layer = self.decoder_layer.model.get_input_embeddings()
        return embedding_layer(input_ids)
    
    @torch.inference_mode()
    def generate_from_embeds_custom(
        self,
        inputs_embeds: torch.Tensor,                 # (B, L0, D)
        attention_mask: Optional[torch.Tensor] = None,
        max_new_tokens: int = 32,
        eos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        temperature: float = 1.0,
        do_sample: bool = False,
        labels_trie: Optional[LabelsTrie] = None,
        **kwargs
    ):
        model = self.decoder_layer.model
        device, (B, L0, _) = inputs_embeds.device, inputs_embeds.shape
        cfg = model.config

        eos_token_id = eos_token_id or cfg.eos_token_id
        pad_token_id = pad_token_id or cfg.pad_token_id or eos_token_id

        # prefix mask
        if attention_mask is None:
            attention_mask = torch.ones(B, L0, dtype=torch.long, device=device)

        out = model(inputs_embeds=inputs_embeds,
                    attention_mask=attention_mask,
                    use_cache=True)
        past_key_values = out.past_key_values
        next_logits = out.logits[:, -1]                         # (B, V)

        unfinished = torch.ones(B, dtype=torch.bool, device=device)
        generated = [[] for _ in range(B)]

        for _ in range(max_new_tokens):
            if labels_trie is not None:
                V = next_logits.shape[1]
                mask_tensor = torch.full((B, V), -float('inf'), device=device)
                for b in range(B):
                    if unfinished[b]:
                        current_seq = generated[b]  # Tokens generated so far
                        
                        allowed_tokens = labels_trie.get(current_seq)

                        if len(allowed_tokens) == 0:
                            allowed_tokens = [eos_token_id]
                        
                        mask_tensor[b, allowed_tokens] = 0
                    else:
                         mask_tensor[b, :] = 0
                next_logits = next_logits + mask_tensor

            if temperature != 1.0:
                next_logits = next_logits / temperature

            if do_sample:
                probs = torch.softmax(next_logits, dim=-1)
                next_token = torch.multinomial(probs, num_samples=1)      # (B, 1)
            
            else:
                next_token = next_logits.argmax(dim=-1, keepdim=True)     # (B, 1)

            for b in range(B):
                if unfinished[b]:
                    generated[b].append(next_token[b, 0].item())

            eos_hit = next_token.squeeze() == eos_token_id
            unfinished = unfinished & ~eos_hit
            if not unfinished.any():         
                break

            next_token = next_token.masked_fill(~unfinished.unsqueeze(1), pad_token_id)

            attention_mask = torch.cat(
                [attention_mask, torch.ones(B, 1, dtype=torch.long, device=device)],
                dim=1,
            )

            out = model(input_ids=next_token,
                        attention_mask=attention_mask,
                        past_key_values=past_key_values,
                        use_cache=True)
            past_key_values = out.past_key_values
            next_logits = out.logits[:, -1]

        max_len = max(len(seq) for seq in generated)
        out_ids = torch.full((B, max_len), pad_token_id, dtype=torch.long, device=device)
        for b, seq in enumerate(generated):
            if seq:
                out_ids[b, :len(seq)] = torch.tensor(seq, device=device)

        return out_ids 
    
    @torch.inference_mode()
    def generate_from_embeds(
        self,
        inputs_embeds: torch.Tensor,                 # (B, L0, D)
        attention_mask: Optional[torch.Tensor] = None,
        max_new_tokens: int = 32,
        eos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        temperature: float = 1.0,
        do_sample: bool = False,
        num_return_sequences=1,
        labels_trie: Optional[LabelsTrie] = None,
        **kwargs
    ):
        model = self.decoder_layer.model
        inputs_embeds = inputs_embeds.to(dtype=model.dtype)
        if attention_mask is not None:
            attention_mask = attention_mask.to(dtype=model.dtype)
        device, (B, L0, _) = inputs_embeds.device, inputs_embeds.shape
        cfg = model.config

        # Set token IDs if not provided
        eos_token_id = eos_token_id or cfg.eos_token_id
        pad_token_id = pad_token_id or cfg.pad_token_id or eos_token_id

        # Create attention mask if not provided
        if attention_mask is None:
            attention_mask = torch.ones(B, L0, dtype=torch.long, device=device)

        # Define prefix-constrained token function if trie is provided
        if labels_trie is not None:
            def prefix_allowed_tokens(batch_idx, input_ids):
                current_seq = input_ids.tolist()
                allowed_tokens = labels_trie.get(current_seq)
                if not allowed_tokens:  # Empty or None
                    allowed_tokens = [eos_token_id]
                return allowed_tokens
        else:
            prefix_allowed_tokens = None

        # Generate new tokens using transformer's generate method
        generated_ids = model.generate(
            inputs_embeds=inputs_embeds, 
            attention_mask=attention_mask,
            past_key_values=None,
            max_new_tokens=max_new_tokens,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            temperature=temperature,
            do_sample=do_sample,
            use_cache=True,
            num_return_sequences=num_return_sequences,
            num_beams=num_return_sequences,
            prefix_allowed_tokens_fn=prefix_allowed_tokens,
            logits_processor=LogitsProcessorList([
                NumericalStabilityProcessor(),
            ]),
            **kwargs
        )

        return generated_ids
        
        
    def generate(self, *args, **kwargs):
        if "inputs_embeds" in kwargs:
            return self.generate_from_embeds(*args, **kwargs)
        else:
            return self.decoder_layer.model.generate(*args, **kwargs) 
    
    def forward(self, *args, **kwargs) -> torch.Tensor:
        decoded_embeddings = self.decoder_layer(*args, **kwargs)
        return decoded_embeddings